﻿using NFC.Crypto;
using NFC.Mifare_DESFire.Enums;
using PCSC.Iso7816;
using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Http;
using System.Text;

namespace NFC.Mifare_DESFire
{

    public class MifareDESFire
    {
        public byte[] GenerateDefaultKey(int size)
        {
            List<byte> key = new List<byte>();
            for (int i = 0; i < size; i++)
            {
                key.Add(0);
            }

            return key.ToArray();
        }

        private ICard _Card;

        public MifareDESFire(ICard card)
        {
            _Card = card;
        }

        /// <summary>
        /// Create new Application with AID
        /// </summary>
        /// <param name="aid">Appilication ID</param>
        public APDUCommand CreateApplication(UInt64 aid)
        {
            throw new NotImplementedException();
        }

        public APDUCommand GetApplicationIDs()
        {
            APDUCommand cmd = new APDUCommand(IsoCase.Case2Short)
            {
                CLA = 0x90,
                INS = (byte)APDUInstructions.GET_APPLICATION_IDS
            };

            return cmd;
        }

        public UInt32[] ConvertApplicationIDs(APDUResponse response)
        {
            if(response.Body.Length % 3 != 0)
            {
                throw new Exception("Invalid Body Length.");
            }

            List<UInt32> applicationIDs = new List<UInt32>();

            for(int i = 0; i < response.Body.Length; i += 3)
            {
                UInt32 new_applicationID = 0;
                new_applicationID = (UInt32)((response.Body[i] << 16) + (response.Body[i + 1] << 8) + response.Body[i + 2]);
                applicationIDs.Add(new_applicationID);
            }

            return applicationIDs.ToArray();
        }

        public void Format()
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Authenticate to Card
        /// </summary>
        /// <param name="key_id">0x01 - 0x0D</param>
        /// <param name="key"></param>
        public void Authenticate(byte key_id, byte[] key)
        {
            APDUCommand cmd_getchallange = new APDUCommand(IsoCase.Case4Short)
            {
                CLA = 0x90,
                INS = (byte)0x1A,
                Data = new byte[]
                {
                    key_id
                }
            };
            APDUResponse response = _Card.Transmit(cmd_getchallange);

            byte[] challenge = response.Body;
            Console.WriteLine("Challange: {0}", toHexString(challenge));

            byte[] rndA = new byte[]
            {
                0x92, 0x31, 0x34, 0x8B, 0x66, 0x35, 0xA8, 0xAF
            };
            Console.WriteLine("rndA: {0}", toHexString(rndA));

            TripleDES des = new TripleDES();
            byte[] rndB = des.Decrypt(challenge, key, GenerateDefaultKey(8));
            Console.WriteLine("rndB: {0}", toHexString(rndB));

            byte[] leftRotatedRndB = rotateLeft(rndB);
            Console.WriteLine("leftRotatedRndB: {0}", toHexString(leftRotatedRndB));

            byte[] rndA_rndB = concatenate(rndA, leftRotatedRndB);
            Console.WriteLine("rndA_rndB: {0}", toHexString(rndA_rndB));

            byte[] challengeAnswer = des.Encrypt(rndA_rndB, key, GenerateDefaultKey(8));
            Console.WriteLine("challengeAnswer: {0}", toHexString(challengeAnswer));

            APDUCommand cmd_answerchallange = new APDUCommand(IsoCase.Case4Short)
            {
                CLA = 0x90,
                INS = (byte)0xAF,
                Data = challengeAnswer
            };
            Console.WriteLine("cmd_answerchallange: {0}", toHexString(cmd_answerchallange.ToArray()));
           
            response = _Card.Transmit(cmd_answerchallange);

            byte[] encryptedRndAFromCard = response.Body;
            Console.WriteLine("encryptedRndAFromCard: {0}", toHexString(encryptedRndAFromCard));

            byte[] rotatedRndAFromCard = des.Decrypt(encryptedRndAFromCard, key, GenerateDefaultKey(8));
            Console.WriteLine("rotatedRndAFromCard: {0}", toHexString(rotatedRndAFromCard));

            byte[] rndAFromCard = rotateRight(rotatedRndAFromCard);
            Console.WriteLine("rndAFromCard: {0}", toHexString(rndAFromCard));

            if (!rndA.Equals(rndAFromCard))
            {
                throw new Exception("???");
            }
        }

        private String toHexString(byte[] data)
        {
            return BitConverter.ToString(data).Replace("-", string.Empty);
        }

        public byte[] rotateLeft(byte[] data)
        {
            byte[] rotate = new byte[data.Length];
            data.CopyTo(rotate, 0);

            byte temp = rotate[0];
            for (var i = 0; i < rotate.Length - 1; i++)
            {
                rotate[i] = rotate[i + 1];
            }
            rotate[rotate.Length - 1] = temp;

            return rotate;
        }

        public byte[] rotateRight(byte[] data)
        {
            byte[] rotate = new byte[data.Length];
            data.CopyTo(rotate, 0);

            byte temp = rotate[rotate.Length - 1];
            for (var i = rotate.Length - 1; i > 1; i--)
            {
                rotate[i] = rotate[i - 1];
            }
            rotate[0] = temp;

            return rotate;
        }

        public byte[] concatenate(byte[] a, byte[] b)
        {
            byte[] c = new byte[a.Length + b.Length];
            a.CopyTo(c, 0);
            b.CopyTo(c, a.Length);

            return c;
        }

        /// <summary>
        /// Select Application by ID
        /// </summary>
        /// <param name="id">3 Byte ID</param>
        public APDUCommand SelectApplication(UInt32 id)
        {
            byte[] id_byte = BitConverter.GetBytes(id);

            APDUCommand cmd = new APDUCommand(IsoCase.Case4Short)
            {
                CLA = 0x90,
                INS = (byte)APDUInstructions.SELECT_APPLICATION,
                Data = new byte[]
                {
                    id_byte[0],
                    id_byte[1],
                    id_byte[2]
                },
                Le = 0x00
            };

            return cmd;
        }

        public void ChangeApplicationMasterKey(byte[] aPP_MasterKey)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Delete Application by ID
        /// </summary>
        /// <param name="id">3 Byte ID</param>
        public APDUCommand DeleteApplication(UInt32 id)
        {
            byte[] id_byte = BitConverter.GetBytes(id);

            APDUCommand cmd = new APDUCommand(IsoCase.Case4Short)
            {
                CLA = 0x90,
                INS = (byte)APDUInstructions.DELETE_APPLICATION,
                Data = new byte[]
                {
                    id_byte[0],
                    id_byte[1],
                    id_byte[2]
                },
                Le = 0x00
            };

            return cmd;
        }

        public void ChangeApplicationKey(int v, byte[] aPP_Key_1)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Select Application by ID
        /// </summary>
        /// <param name="id">3 Byte ID</param>
        public APDUCommand CreateApplication(UInt32 id, byte keysetting1, byte keysetting2)
        {
            byte[] id_byte = BitConverter.GetBytes(id);

            APDUCommand cmd = new APDUCommand(IsoCase.Case4Short)
            {
                CLA = 0x90,
                INS = (byte)APDUInstructions.CREATE_APPLICATION,
                Data = new byte[]
                {
                    id_byte[0],
                    id_byte[1],
                    id_byte[2],
                    keysetting1,
                    keysetting2
                },
                Le = 0x00
            };

            return cmd;
        }

        public void CreateFile(byte fabAccessIdentFileID, FileCommunication pLAIN, ushort fileAccessRight, UInt32 v)
        {
            throw new NotImplementedException();
        }

        public ushort GenerateFileAccessRight(AccessRights fREE, int v1, int v2, int v3)
        {
            throw new NotImplementedException();
        }

        public void WirteData(byte fabAccessIdentFileID, int v1, int v2, byte[] vs)
        {
            throw new NotImplementedException();
        }

        public void CreateFile(byte fabAccessIdentFileID, object plain, AccessRights fREE, int v1, int v2, int v3)
        {
            throw new NotImplementedException();
        }

        public byte[] ReadData(byte identFileID, int v1, int v2)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Genearte KeySetting1 for Application Settings or PICC Setting
        /// </summary>
        public byte GenerateKeySetting1(ChangeApplicationKey changeKey, ChangeMasterKeySettings changeMasterKeySettings, CreateDeleteFile createDeleteFile, FileDirectoryAccess fileDirectoryAccess, ChangeMasterKey changeMasterKey)
        {
            return (byte)(((byte)changeKey << 4) | (byte)changeMasterKeySettings | (byte)createDeleteFile | (byte)fileDirectoryAccess | (byte)changeMasterKey);
        }

        /// <summary>
        /// Genearte KeySetting1 for Application Settings or PICC Setting
        /// </summary>
        /// <param name="changeKey">ID of Key for changing Application Keys</param>
        /// <returns></returns>
        public byte GenerateKeySetting1(byte changeKey, ChangeMasterKeySettings changeMasterKeySettings, CreateDeleteFile createDeleteFile, FileDirectoryAccess fileDirectoryAccess, ChangeMasterKey changeMasterKey)
        {
            if(changeKey < 0x01 || changeKey >= 0x0E)
            {
                throw new ArgumentOutOfRangeException();
            }
            return GenerateKeySetting1(changeKey, changeMasterKeySettings, createDeleteFile, fileDirectoryAccess, changeMasterKey);
        }

        /// <summary>
        /// Genearte KeySetting2 for Application Creation
        /// </summary>
        /// <param name="numberOfKeys">Number of keys that can be stored within the application (0x01-0x0D)</param>
        /// <returns></returns>
        public byte GenerateKeySetting2(CryptoOperationsType cryptoOperations, FileIdentifies fileIdentifies, byte numberOfKeys)
        {
            if(numberOfKeys < 0x01 || numberOfKeys >= 0x0D)
            {
                throw new ArgumentOutOfRangeException();
            }

            return (byte)((byte)cryptoOperations | (byte)fileIdentifies | numberOfKeys);
        }
    }
}
